@@ -1,20 +1,20 @@
-use core::{convert::TryInto, ptr, time::Duration};
-
-extern crate alloc;
-use alloc::boxed::Box;
-use alloc::sync::Arc;
+use core::cell::UnsafeCell;
+use core::fmt::Debug;
+use core::ptr;
+use core::time::Duration;
 
 use ::log::*;
 
 use enumset::*;
 
+extern crate alloc;
+use alloc::sync::Arc;
+
+use embedded_svc::errors::Errors;
 use embedded_svc::eth::*;
+use embedded_svc::event_bus::EventBus;
 use embedded_svc::ipv4;
 
-use esp_idf_sys::*;
-
-use crate::private::waitable::*;
-
 #[cfg(any(
     all(esp32, esp_idf_eth_use_esp32_emac),
     any(
@@ -24,7 +24,6 @@ use crate::private::waitable::*;
     )
 ))]
 use esp_idf_hal::gpio;
-
 #[cfg(any(
     esp_idf_eth_spi_ethernet_dm9051,
     esp_idf_eth_spi_ethernet_w5500,
@@ -32,10 +31,16 @@ use esp_idf_hal::gpio;
 ))]
 use esp_idf_hal::{spi, units::Hertz};
 
+use esp_idf_sys::*;
+
+use crate::eventloop::{EspSubscription, EspTypedEventDeserializer, EspTypedEventSource, System};
 use crate::netif::*;
+use crate::private::common::UnsafeCellSendSync;
+use crate::private::waitable::*;
 use crate::sysloop::*;
 
-use crate::private::common::*;
+#[cfg(feature = "experimental")]
+pub use nonblocking::*;
 
 #[cfg(all(esp32, esp_idf_eth_use_esp32_emac))]
 // TODO: #[derive(Debug)]
@@ -114,7 +119,7 @@ struct Shared {
     operating: bool,
 
     handle: esp_eth_handle_t,
-    netif: Option<*mut esp_netif_t>,
+    netif: Option<EspNetif>,
 }
 
 impl Shared {
@@ -128,22 +133,31 @@ impl Shared {
             netif: None,
         }
     }
+
+    fn is_our_eth_event(&self, event: &EthEvent) -> bool {
+        self.handle as *const _ == event.handle()
+    }
+
+    fn is_our_ip_event(&self, event: &IpEvent) -> bool {
+        self.netif.is_some()
+            && self.netif.as_ref().map(|netif| netif.1) == event.handle().map(|handle| handle as _)
+    }
 }
 
 unsafe impl Send for Shared {}
 
 pub struct EspEth<P> {
     netif_stack: Arc<EspNetifStack>,
-    _sys_loop_stack: Arc<EspSysLoopStack>,
+    sys_loop_stack: Arc<EspSysLoopStack>,
 
     peripherals: P,
 
-    handle: esp_eth_handle_t,
     glue_handle: *mut c_types::c_void,
 
-    netif: Option<EspNetif>,
+    waitable: Arc<Waitable<Shared>>,
 
-    shared: Box<Waitable<Shared>>,
+    _eth_subscription: EspSubscription<System>,
+    _ip_subscription: EspSubscription<System>,
 }
 
 #[cfg(all(esp32, esp_idf_eth_use_esp32_emac))]
@@ -490,37 +504,44 @@ impl<P> EspEth<P> {
 
         let glue_handle = unsafe { esp_eth_new_netif_glue(handle) };
 
-        let mut shared: Box<Waitable<Shared>> = Box::new(Waitable::new(Shared::new(handle)));
+        let waitable: Arc<Waitable<Shared>> = Arc::new(Waitable::new(Shared::new(handle)));
 
-        let shared_ref: *mut _ = &mut *shared;
+        let eth_waitable = waitable.clone();
+        let eth_subscription =
+            sys_loop_stack
+                .get_loop()
+                .clone()
+                .subscribe(move |event: &EthEvent| {
+                    let mut shared = eth_waitable.state.lock();
 
-        esp!(unsafe {
-            esp_event_handler_register(
-                ETH_EVENT,
-                ESP_EVENT_ANY_ID,
-                Option::Some(Self::event_handler),
-                shared_ref as *mut c_types::c_void,
-            )
-        })?;
-        esp!(unsafe {
-            esp_event_handler_register(
-                IP_EVENT,
-                ESP_EVENT_ANY_ID,
-                Option::Some(Self::event_handler),
-                shared_ref as *mut c_types::c_void,
-            )
-        })?;
+                    if Self::on_eth_event(&mut shared, event).unwrap() {
+                        eth_waitable.cvar.notify_all();
+                    }
+                })?;
+
+        let ip_waitable = waitable.clone();
+        let ip_subscription =
+            sys_loop_stack
+                .get_loop()
+                .clone()
+                .subscribe(move |event: &IpEvent| {
+                    let mut shared = ip_waitable.state.lock();
+
+                    if Self::on_ip_event(&mut shared, event).unwrap() {
+                        ip_waitable.cvar.notify_all();
+                    }
+                })?;
 
         info!("Event handlers registered");
 
         let eth = Self {
             netif_stack,
-            _sys_loop_stack: sys_loop_stack,
+            sys_loop_stack,
             peripherals,
-            handle,
             glue_handle: glue_handle as *mut _,
-            netif: None,
-            shared,
+            waitable,
+            _eth_subscription: eth_subscription,
+            _ip_subscription: ip_subscription,
         };
 
         info!("Initialization complete");
@@ -532,18 +553,21 @@ impl<P> EspEth<P> {
     where
         F: FnOnce(Option<&EspNetif>) -> T,
     {
-        f(self.netif.as_ref())
+        self.waitable.get(|shared| f(shared.netif.as_ref()))
     }
 
     pub fn with_netif_mut<F, T>(&mut self, f: F) -> T
     where
         F: FnOnce(Option<&mut EspNetif>) -> T,
     {
-        f(self.netif.as_mut())
+        self.waitable.get_mut(|shared| f(shared.netif.as_mut()))
     }
 
     fn set_ip_conf(&mut self, conf: &Configuration) -> Result<(), EspError> {
-        Self::netif_unbind(self.netif.as_mut())?;
+        {
+            let mut shared = self.waitable.state.lock();
+            Self::netif_unbind(shared.netif.as_mut())?;
+        }
 
         let iconf = match conf {
             Configuration::Client(conf) => {
@@ -565,48 +589,45 @@ impl<P> EspEth<P> {
             _ => None,
         };
 
-        if let Some(iconf) = iconf {
+        let netif = if let Some(iconf) = iconf {
             let netif = EspNetif::new(self.netif_stack.clone(), &iconf)?;
 
             esp!(unsafe { esp_netif_attach(netif.1, self.glue_handle) })?;
 
-            self.netif = Some(netif);
-
             info!("IP configuration done");
+
+            Some(netif)
         } else {
-            self.netif = None;
-
             info!("Skipping IP configuration (not configured)");
-        }
 
-        let netif_ptr = self.netif.as_ref().map(|netif| netif.1);
+            None
+        };
 
-        self.shared.modify(|shared| {
+        {
+            let mut shared = self.waitable.state.lock();
             shared.conf = conf.clone();
-            shared.netif = netif_ptr;
-
-            (false, ())
-        });
+            shared.netif = netif;
+        }
 
         Ok(())
     }
 
-    fn wait_status(&self, matcher: impl Fn(&Status) -> bool) {
+    pub fn wait_status(&self, matcher: impl Fn(&Status) -> bool) {
         info!("About to wait for status");
 
-        self.shared.wait_while(|shared| !matcher(&shared.status));
+        self.waitable.wait_while(|shared| !matcher(&shared.status));
 
         info!("Waiting for status done - success");
     }
 
-    fn wait_status_with_timeout(
+    pub fn wait_status_with_timeout(
         &self,
         dur: Duration,
         matcher: impl Fn(&Status) -> bool,
     ) -> Result<(), Status> {
         info!("About to wait {:?} for status", dur);
 
-        let (timeout, status) = self.shared.wait_timeout_while_and_get(
+        let (timeout, status) = self.waitable.wait_timeout_while_and_get(
             dur,
             |shared| !matcher(&shared.status),
             |shared| shared.status.clone(),
@@ -621,25 +642,30 @@ impl<P> EspEth<P> {
         }
     }
 
-    fn start(&mut self, status: Status) -> Result<(), EspError> {
+    fn start(&mut self, status: Status, wait: Option<Duration>) -> Result<(), EspError> {
         info!("Starting with status: {:?}", status);
 
-        self.shared.modify(|shared| {
+        {
+            let mut shared = self.waitable.state.lock();
+
             shared.status = status.clone();
             shared.operating = shared.status.is_operating();
 
-            (false, ())
-        });
+            if status.is_operating() {
+                info!("Status is of operating type, starting");
 
-        if status.is_operating() {
-            info!("Status is of operating type, starting");
+                esp!(unsafe { esp_eth_start(shared.handle) })?;
 
-            esp!(unsafe { esp_eth_start(self.handle) })?;
+                info!("Start requested");
 
-            info!("Start requested");
+                Self::netif_info("ETH", shared.netif.as_ref())?;
+            } else {
+                info!("Status is NOT of operating type, not starting");
+            }
+        }
 
-            let result =
-                self.wait_status_with_timeout(Duration::from_secs(10), |s| !s.is_transitional());
+        if let Some(duration) = wait {
+            let result = self.wait_status_with_timeout(duration, |s| !s.is_transitional());
 
             if result.is_err() {
                 info!("Timeout while waiting for the requested state");
@@ -648,64 +674,48 @@ impl<P> EspEth<P> {
             }
 
             info!("Started");
-
-            Self::netif_info("ETH", self.netif.as_ref())?;
-        } else {
-            info!("Status is NOT of operating type, not starting");
         }
 
         Ok(())
     }
 
-    fn stop(&mut self) -> Result<(), EspError> {
+    fn stop(&mut self, wait: bool) -> Result<(), EspError> {
         info!("Stopping");
 
-        self.shared.modify(|shared| {
+        {
+            let mut shared = self.waitable.state.lock();
+
             shared.operating = false;
 
-            (false, ())
-        });
+            let err = unsafe { esp_eth_stop(shared.handle) };
+            if err != ESP_ERR_INVALID_STATE as i32 {
+                esp!(err)?;
+            }
 
-        let err = unsafe { esp_eth_stop(self.handle) };
-        if err != ESP_ERR_INVALID_STATE as i32 {
-            esp!(err)?;
+            info!("Stop requested");
         }
-        info!("Stop requested");
 
-        self.wait_status(|s| matches!(s, Status::Stopped));
-
-        info!("Stopped");
+        if wait {
+            self.wait_status(|s| matches!(s, Status::Stopped));
+        }
 
         Ok(())
     }
 
     fn clear_all(&mut self) -> Result<(), EspError> {
-        self.stop()?;
+        self.stop(true)?;
+
+        let mut shared = self.waitable.state.lock();
 
         unsafe {
-            Self::netif_unbind(self.netif.as_mut())?;
-            self.shared.modify(|shared| {
-                shared.netif = None;
-
-                (false, ())
-            });
+            Self::netif_unbind(shared.netif.as_mut())?;
+            shared.netif = None;
 
             esp!(esp_eth_del_netif_glue(self.glue_handle as *mut _))?;
 
-            esp!(esp_event_handler_unregister(
-                ETH_EVENT,
-                ESP_EVENT_ANY_ID,
-                Option::Some(Self::event_handler)
-            ))?;
-            esp!(esp_event_handler_unregister(
-                IP_EVENT,
-                ESP_EVENT_ANY_ID as i32,
-                Option::Some(Self::event_handler)
-            ))?;
-
             info!("Event handlers deregistered");
 
-            esp!(esp_eth_driver_uninstall(self.handle))?;
+            esp!(esp_eth_driver_uninstall(shared.handle))?;
 
             info!("Driver deinitialized");
         }
@@ -736,130 +746,69 @@ impl<P> EspEth<P> {
         Ok(())
     }
 
-    unsafe extern "C" fn event_handler(
-        arg: *mut c_types::c_void,
-        event_base: esp_event_base_t,
-        event_id: c_types::c_int,
-        event_data: *mut c_types::c_void,
-    ) {
-        let shared_ref = (arg as *mut Waitable<Shared>).as_mut().unwrap();
+    fn on_eth_event(shared: &mut Shared, event: &EthEvent) -> Result<bool, EspError> {
+        if shared.is_our_eth_event(event) {
+            info!("Got eth event: {:?} ", event);
 
-        shared_ref.modify(|shared| {
-            if event_base == ETH_EVENT {
-                Self::on_eth_event(shared, event_id, event_data)
-            } else if event_base == IP_EVENT {
-                Self::on_ip_event(shared, event_id, event_data)
+            shared.status = match event {
+                EthEvent::Started(_) => Status::Starting,
+                EthEvent::Stopped(_) => Status::Stopped,
+                EthEvent::Connected(_) => {
+                    Status::Started(ConnectionStatus::Connected(match shared.conf {
+                        Configuration::Client(ipv4::ClientConfiguration::DHCP(_)) => {
+                            IpStatus::Waiting
+                        }
+                        Configuration::Client(ipv4::ClientConfiguration::Fixed(ref status)) => {
+                            IpStatus::Done(Some(status.clone()))
+                        }
+                        Configuration::Router(_) => IpStatus::Done(None),
+                        _ => IpStatus::Disabled,
+                    }))
+                }
+                EthEvent::Disconnected(_) => Status::Started(ConnectionStatus::Disconnected),
+            };
+
+            info!(
+                "Eth event {:?} handled, set status: {:?}",
+                event, shared.status
+            );
+
+            Ok(true)
+        } else {
+            Ok(false)
+        }
+    }
+
+    fn on_ip_event(shared: &mut Shared, event: &IpEvent) -> Result<bool, EspError> {
+        if shared.is_our_ip_event(event) {
+            info!("Got IP event: {:?}", event);
+
+            let status = match event {
+                IpEvent::DhcpIpAssigned(assignment) => Some(Status::Started(
+                    ConnectionStatus::Connected(IpStatus::Done(Some(assignment.ip_settings))),
+                )),
+                IpEvent::DhcpIpDeassigned(_) => Some(Status::Started(ConnectionStatus::Connected(
+                    IpStatus::Waiting,
+                ))),
+                _ => None,
+            };
+
+            if let Some(status) = status {
+                shared.status = status;
+                info!(
+                    "IP event {:?} handled, set status: {:?}",
+                    event, shared.status
+                );
+
+                Ok(true)
             } else {
-                warn!("Got unknown event base");
+                info!("IP event {:?} skipped", event);
 
                 Ok(false)
             }
-            .map(|notify| (notify, ()))
-            .unwrap()
-        });
-    }
-
-    #[allow(non_upper_case_globals)]
-    fn on_eth_event(
-        shared: &mut Shared,
-        event_id: c_types::c_int,
-        event_data: *mut c_types::c_void,
-    ) -> Result<bool, EspError> {
-        let eth_handle = unsafe { (event_data as *const esp_eth_handle_t).as_ref() };
-        let for_us = eth_handle
-            .map(|eth_handle| *eth_handle == shared.handle)
-            .unwrap_or(false);
-
-        if !for_us {
-            return Ok(false);
+        } else {
+            Ok(false)
         }
-
-        info!("Got eth event: {} ", event_id);
-
-        let handled = match event_id as u32 {
-            eth_event_t_ETHERNET_EVENT_START => {
-                shared.status = Status::Starting;
-                true
-            }
-            eth_event_t_ETHERNET_EVENT_STOP => {
-                shared.status = Status::Stopped;
-                true
-            }
-            eth_event_t_ETHERNET_EVENT_CONNECTED => {
-                shared.status = Status::Started(ConnectionStatus::Connected(match shared.conf {
-                    Configuration::Client(ipv4::ClientConfiguration::DHCP(_)) => IpStatus::Waiting,
-                    Configuration::Client(ipv4::ClientConfiguration::Fixed(ref status)) => {
-                        IpStatus::Done(Some(status.clone()))
-                    }
-                    Configuration::Router(_) => IpStatus::Done(None),
-                    _ => IpStatus::Disabled,
-                }));
-
-                true
-            }
-            eth_event_t_ETHERNET_EVENT_DISCONNECTED => {
-                shared.status = Status::Started(ConnectionStatus::Disconnected);
-
-                true
-            }
-            _ => false,
-        };
-
-        if handled {
-            info!(
-                "Eth event {} handled, set status: {:?}",
-                event_id, shared.status
-            );
-        }
-
-        Ok(handled)
-    }
-
-    #[allow(non_upper_case_globals)]
-    fn on_ip_event(
-        shared: &mut Shared,
-        event_id: c_types::c_int,
-        event_data: *mut c_types::c_void,
-    ) -> Result<bool, EspError> {
-        let event_id = event_id as u32;
-
-        let for_us = shared.netif.is_some()
-            && (event_id == ip_event_t_IP_EVENT_ETH_GOT_IP
-                || event_id == ip_event_t_IP_EVENT_STA_GOT_IP);
-        if !for_us {
-            return Ok(false);
-        }
-
-        let event = unsafe { (event_data as *const ip_event_got_ip_t).as_ref() };
-        if !event.is_some() {
-            return Ok(false);
-        }
-
-        let event = event.unwrap();
-        if event.esp_netif != shared.netif.unwrap() {
-            return Ok(false);
-        }
-
-        info!("Got IP event: {}", event_id);
-
-        shared.status = Status::Started(ConnectionStatus::Connected(IpStatus::Done(Some(
-            ipv4::ClientSettings {
-                ip: ipv4::Ipv4Addr::from(Newtype(event.ip_info.ip)),
-                subnet: ipv4::Subnet {
-                    gateway: ipv4::Ipv4Addr::from(Newtype(event.ip_info.gw)),
-                    mask: Newtype(event.ip_info.netmask).try_into()?,
-                },
-                dns: None,           // TODO
-                secondary_dns: None, // TODO
-            },
-        ))));
-
-        info!(
-            "IP event {} handled, set status: {:?}",
-            event_id, shared.status
-        );
-
-        Ok(true)
     }
 
     fn eth_default_config(mac: *mut esp_eth_mac_t, phy: *mut esp_eth_phy_t) -> esp_eth_config_t {
@@ -903,9 +852,11 @@ impl<P> EspEth<P> {
     }
 }
 
-impl<P> Eth for EspEth<P> {
+impl<P> Errors for EspEth<P> {
     type Error = EspError;
+}
 
+impl<P> Eth for EspEth<P> {
     fn get_capabilities(&self) -> Result<EnumSet<Capability>, Self::Error> {
         let caps = Capability::Client | Capability::Router;
 
@@ -915,7 +866,7 @@ impl<P> Eth for EspEth<P> {
     }
 
     fn get_status(&self) -> Status {
-        let status = self.shared.get(|shared| shared.status.clone());
+        let status = self.waitable.get(|shared| shared.status.clone());
 
         info!("Providing status: {:?}", status);
 
@@ -926,7 +877,7 @@ impl<P> Eth for EspEth<P> {
     fn get_configuration(&self) -> Result<Configuration, Self::Error> {
         info!("Getting configuration");
 
-        let conf = self.shared.get(|shared| shared.conf.clone());
+        let conf = self.waitable.get(|shared| shared.conf.clone());
 
         info!("Configuration gotten: {:?}", &conf);
 
@@ -936,7 +887,7 @@ impl<P> Eth for EspEth<P> {
     fn set_configuration(&mut self, conf: &Configuration) -> Result<(), Self::Error> {
         info!("Setting configuration: {:?}", conf);
 
-        self.stop()?;
+        self.stop(false)?;
 
         self.set_ip_conf(conf)?;
 
@@ -946,10 +897,153 @@ impl<P> Eth for EspEth<P> {
             Status::Starting
         };
 
-        self.start(status)?;
+        self.start(status, None)?;
 
         info!("Configuration set");
 
         Ok(())
     }
 }
+
+pub type EthHandle = *const core::ffi::c_void;
+
+#[derive(Copy, Clone, Debug, Eq, PartialEq)]
+pub enum EthEvent {
+    Started(EthHandle),
+    Stopped(EthHandle),
+    Connected(EthHandle),
+    Disconnected(EthHandle),
+}
+
+impl EthEvent {
+    pub fn handle(&self) -> EthHandle {
+        match self {
+            Self::Started(handle) => *handle,
+            Self::Stopped(handle) => *handle,
+            Self::Connected(handle) => *handle,
+            Self::Disconnected(handle) => *handle,
+        }
+    }
+}
+
+impl EspTypedEventSource for EthEvent {
+    fn source() -> *const c_types::c_char {
+        unsafe { ETH_EVENT }
+    }
+}
+
+impl EspTypedEventDeserializer<EthEvent> for EthEvent {
+    #[allow(non_upper_case_globals, non_snake_case)]
+    fn deserialize<R>(
+        data: &crate::eventloop::EspEventFetchData,
+        f: &mut impl for<'a> FnMut(&'a EthEvent) -> R,
+    ) -> R {
+        let eth_handle_ref = unsafe { (data.payload as *const esp_eth_handle_t).as_ref() };
+
+        let event_id = data.event_id as u32;
+
+        let event = if event_id == eth_event_t_ETHERNET_EVENT_START {
+            EthEvent::Started(*eth_handle_ref.unwrap() as _)
+        } else if event_id == eth_event_t_ETHERNET_EVENT_STOP {
+            EthEvent::Stopped(*eth_handle_ref.unwrap() as _)
+        } else if event_id == eth_event_t_ETHERNET_EVENT_CONNECTED {
+            EthEvent::Connected(*eth_handle_ref.unwrap() as _)
+        } else if event_id == eth_event_t_ETHERNET_EVENT_DISCONNECTED {
+            EthEvent::Disconnected(*eth_handle_ref.unwrap() as _)
+        } else {
+            panic!("Unknown event ID: {}", event_id);
+        };
+
+        f(&event)
+    }
+}
+
+impl<P> EventBus<()> for EspEth<P> {
+    type Subscription = (EspSubscription<System>, EspSubscription<System>);
+
+    fn subscribe(
+        &mut self,
+        callback: impl for<'a> FnMut(&'a ()) + Send + 'static,
+    ) -> Result<Self::Subscription, Self::Error> {
+        let eth_cb = Arc::new(UnsafeCellSendSync(UnsafeCell::new(callback)));
+        let eth_last_status = Arc::new(UnsafeCellSendSync(UnsafeCell::new(self.get_status())));
+        let eth_waitable = self.waitable.clone();
+
+        let ip_cb = eth_cb.clone();
+        let ip_last_status = eth_last_status.clone();
+        let ip_waitable = eth_waitable.clone();
+
+        let subscription1 =
+            self.sys_loop_stack
+                .get_loop()
+                .clone()
+                .subscribe(move |event: &EthEvent| {
+                    let notify = {
+                        let shared = eth_waitable.state.lock();
+
+                        if shared.is_our_eth_event(event) {
+                            let last_status_ref =
+                                unsafe { eth_last_status.0.get().as_mut().unwrap() };
+
+                            if *last_status_ref != shared.status {
+                                *last_status_ref = shared.status.clone();
+
+                                true
+                            } else {
+                                false
+                            }
+                        } else {
+                            false
+                        }
+                    };
+
+                    if notify {
+                        let cb_ref = unsafe { eth_cb.0.get().as_mut().unwrap() };
+
+                        (cb_ref)(&());
+                    }
+                })?;
+
+        let subscription2 =
+            self.sys_loop_stack
+                .get_loop()
+                .clone()
+                .subscribe(move |event: &IpEvent| {
+                    let notify = {
+                        let shared = ip_waitable.state.lock();
+
+                        if shared.is_our_ip_event(event) {
+                            let last_status_ref =
+                                unsafe { ip_last_status.0.get().as_mut().unwrap() };
+
+                            if *last_status_ref != shared.status {
+                                *last_status_ref = shared.status.clone();
+
+                                true
+                            } else {
+                                false
+                            }
+                        } else {
+                            false
+                        }
+                    };
+
+                    if notify {
+                        let cb_ref = unsafe { ip_cb.0.get().as_mut().unwrap() };
+
+                        (cb_ref)(&());
+                    }
+                })?;
+
+        Ok((subscription1, subscription2))
+    }
+}
+
+#[cfg(feature = "experimental")]
+mod nonblocking {
+    use embedded_svc::utils::nonblocking::{event_bus::Channel, Asyncify};
+
+    impl<P> Asyncify for super::EspEth<P> {
+        type AsyncWrapper<S> = Channel<(), esp_idf_hal::mutex::Condvar, S>;
+    }
+}
