@@ -1,12 +1,17 @@
 #![cfg_attr(not(feature = "std"), no_std)]
-#![cfg_attr(feature = "experimental", feature(generic_associated_types))] // for http, http::client, http::server, ota
+#![feature(generic_associated_types)] // For mutex, http, http::client, http::server, ota
 #![feature(const_btree_new)]
 
 #[cfg(any(feature = "alloc"))]
 #[macro_use]
 extern crate alloc;
 
-#[cfg(feature = "alloc")]
+#[cfg(all(
+    feature = "alloc",
+    esp_idf_comp_esp_eth_enabled,
+    esp_idf_comp_esp_event_enabled,
+    esp_idf_comp_esp_netif_enabled
+))]
 #[cfg(any(
     all(esp32, esp_idf_eth_use_esp32_emac),
     any(
@@ -17,7 +22,7 @@ extern crate alloc;
     esp_idf_eth_use_openeth
 ))]
 pub mod eth;
-#[cfg(all(feature = "experimental", feature = "alloc"))]
+#[cfg(all(feature = "alloc", esp_idf_comp_esp_event_enabled))]
 pub mod eventloop;
 #[cfg(all(feature = "experimental", feature = "alloc"))]
 pub mod http;
@@ -27,11 +32,15 @@ pub mod httpd;
 #[cfg(feature = "alloc")]
 // TODO: Ideally should not need "alloc" (also for performance reasons)
 pub mod log;
-#[cfg(all(feature = "experimental", feature = "alloc"))]
+#[cfg(all(
+    feature = "alloc",
+    esp_idf_comp_mqtt_enabled,
+    esp_idf_comp_esp_event_enabled
+))]
 pub mod mqtt;
 #[cfg(esp_idf_config_lwip_ipv4_napt)]
 pub mod napt;
-#[cfg(feature = "alloc")]
+#[cfg(all(feature = "alloc", esp_idf_comp_esp_netif_enabled))]
 pub mod netif;
 #[cfg(all(feature = "alloc", esp_idf_comp_nvs_flash_enabled))]
 // TODO: Expose a subset which does not require "alloc"
@@ -49,11 +58,17 @@ pub mod ota;
 pub mod ping;
 #[cfg(feature = "alloc")]
 pub mod sntp;
+#[cfg(esp_idf_comp_esp_event_enabled)]
 pub mod sysloop;
 pub mod systime;
-#[cfg(all(feature = "experimental", feature = "alloc"))]
+#[cfg(all(feature = "alloc", esp_idf_comp_esp_timer_enabled))]
 pub mod timer;
-#[cfg(feature = "alloc")] // TODO: Expose a subset which does not require "alloc"
+#[cfg(all(
+    feature = "alloc",
+    esp_idf_comp_esp_wifi_enabled,
+    esp_idf_comp_esp_event_enabled,
+    esp_idf_comp_esp_netif_enabled
+))]
 pub mod wifi;
 
 mod private;
